from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.contrib.auth.models import Group
from django.contrib.auth import get_user_model
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import get_model
from djinn_auth.models import LocalRole, GlobalRole, Role


def get_group_model():

    """Even though Django doesn't let you override the group model, we
    do...

    """

    group_model = Group

    if getattr(settings, 'AUTH_GROUP_MODEL', None):
        try:
            app_label, model_name = settings.AUTH_GROUP_MODEL.split('.')
        except ValueError:
            raise ImproperlyConfigured(
                "AUTH_GROUP_MODEL must be of the form 'app_label.model_name'")

        group_model = get_model(app_label, model_name)

        if group_model is None:
            raise ImproperlyConfigured("AUTH_GROUP_MODEL refers to model '%s' "
                                       "that has not been installed" %
                                       settings.AUTH_GROUP_MODEL)
    return group_model


def set_local_role(assignee, instance, role):

    """Set the local role. Existing local roles on the given instance with
    the same role will be discarded.

    """

    if type(role) in [str, unicode]:
        role = Role.objects.get(name=role)

    ctype = ContentType.objects.get_for_model(instance)

    LocalRole.objects.filter(
        role=role,
        instance_id=instance.id,
        instance_ct=ctype).delete()

    LocalRole.objects.create(instance=instance, assignee=assignee,
                             role=role)


def assign_local_role(assignee, instance, role):

    """Assign the local role to the given assignee on the instance if it's
    not already there

    """

    instance_ct = ContentType.objects.get_for_model(instance)
    assignee_ct = ContentType.objects.get_for_model(assignee)

    if type(role) in [str, unicode]:
        role = Role.objects.get(name=role)

    LocalRole.objects.get_or_create(instance_ct=instance_ct,
                                    assignee_ct=assignee_ct,
                                    instance_id=instance.id,
                                    assignee_id=assignee.id,
                                    role=role)


def unassign_local_role(assignee, instance, role):

    """Unassign the local role on the given instance for the assignee"""

    instance_ct = ContentType.objects.get_for_model(instance)
    assignee_ct = ContentType.objects.get_for_model(assignee)

    if type(role) in [str, unicode]:
        role = Role.objects.get(name=role)

    LocalRole.objects.filter(instance_ct=instance_ct,
                             assignee_ct=assignee_ct,
                             instance_id=instance.id,
                             assignee_id=assignee.id,
                             role=role).delete()


def assign_global_role(assignee, role):

    """Assign the global role to the assignee if it's not already there"""

    assignee_ct = ContentType.objects.get_for_model(assignee)

    if type(role) in [str, unicode]:
        role = Role.objects.get(name=role)

    GlobalRole.objects.get_or_create(assignee_ct=assignee_ct,
                                     assignee_id=assignee.id,
                                     role=role)


def unassign_global_role(assignee, role):

    """Remove the global role for the asignee"""

    assignee_ct = ContentType.objects.get_for_model(assignee)

    if type(role) in [str, unicode]:
        role = Role.objects.get(name=role)

    GlobalRole.objects.filter(assignee_ct=assignee_ct,
                              assignee_id=assignee.id,
                              role=role).delete()


def get_global_roles(assignee, as_role=False):

    """Return all global roles for the given assignee. If as_role is
    True, return the actual Role objects instead of the GlobalRole
    objects.

    """

    assignee_ct = ContentType.objects.get_for_model(assignee)

    roles = GlobalRole.objects.filter(assignee_ct=assignee_ct,
                                      assignee_id=assignee.id)

    if as_role:
        roles = [role.role for role in roles]

    return roles


def has_global_role(assignee, role):

    assignee_ct = ContentType.objects.get_for_model(assignee)

    if isinstance(role, Role):
        role_name = role.name
    else:
        role_name = role

    return GlobalRole.objects.filter(
        role__name=role_name,
        assignee_id=assignee.id,
        assignee_ct=assignee_ct).exists()


def get_user_global_roles(user, as_role=False):

    """ Get global roles for user, also taking groups into account. If as_role
    is True, return Role instances instead of GlobalRole objects."""

    user_ct = ContentType.objects.get_for_model(get_user_model())
    group_ct = ContentType.objects.get_for_model(get_group_model())

    user_group_ids = user.groups.all().values_list('id', flat=True)

    roles = GlobalRole.objects.filter(
        Q(assignee_ct=user_ct, assignee_id=user.id) |
        Q(assignee_ct=group_ct, assignee_id__in=user_group_ids)
    )

    if as_role:
        roles = [role.role for role in roles]

    return roles


def get_user_local_roles(user, instance, as_role=False):

    """ Get local roles for user, also taking groups into account. If as_role
    is True, return the Role objects instead of the LocalRole objects."""

    user_ct = ContentType.objects.get_for_model(get_user_model())
    group_ct = ContentType.objects.get_for_model(get_group_model())
    instance_ct = ContentType.objects.get_for_model(instance)

    user_group_ids = user.groups.all().values_list('id', flat=True)

    roles = LocalRole.objects.filter(
        Q(instance_ct=instance_ct, instance_id=instance.id,
          assignee_ct=user_ct, assignee_id=user.id) |
        Q(instance_ct=instance_ct, instance_id=instance.id,
          assignee_ct=group_ct, assignee_id__in=user_group_ids)
    )

    if as_role:
        roles = [role.role for role in roles]

    return roles


def get_local_roles(instance, role=None):

    """ Return all local roles on the given instance. If role is set, return
    only local roles that have that role set."""

    ctype = ContentType.objects.get_for_model(instance)

    _filter = {'instance_id': instance.id, 'instance_ct': ctype}

    if role:
        if type(role) in [str, unicode]:
            _filter['role__name'] = role
        else:
            _filter['role'] = role

    return LocalRole.objects.filter(**_filter)


def has_local_role(assignee, instance, role):

    """ Check whether the assignee has the local role """

    instance_ct = ContentType.objects.get_for_model(instance)
    assignee_ct = ContentType.objects.get_for_model(assignee)

    if isinstance(role, Role):
        role_name = role.name
    else:
        role_name = role

    return LocalRole.objects.filter(
        role__name=role_name,
        assignee_id=assignee.id,
        assignee_ct=assignee_ct,
        instance_id=instance.id,
        instance_ct=instance_ct).exists()


def has_user_local_role(user, instance, role):

    """Check whether the user has the local role. This is true if either
    the user directly has the role, or if one of the user groups has
    it.

    """

    if type(role) in [str, unicode]:
        role = Role.objects.get(name=role)

    return get_user_local_roles(user, instance).filter(role=role).exists()



from django.db import models
from django.contrib.contenttypes.models import ContentType
from djinn_auth.models.base import RoleAssignment
try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except:
    # django <1.8 compatibility
    from django.contrib.contenttypes.generic import GenericForeignKey


class LocalRole(RoleAssignment):

    """ Local role for given model instance. Can be assigned either to
    user or to usergroup.
    """

    instance_ct = models.ForeignKey(ContentType, related_name='+')
    instance_id = models.PositiveIntegerField()
    instance = GenericForeignKey('instance_ct', 'instance_id')

    class Meta:

        app_label = "djinn_auth"

    def __unicode__(self):

        return u"%s is %s for %s" % (self.assignee, self.role, self.instance)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from djinn_auth.models.role import Role
try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except:
    # django <1.8 compatibility
    from django.contrib.contenttypes.generic import GenericForeignKey


class RoleAssignment(models.Model):

    """Abstract role assignment class"""

    assignee_ct = models.ForeignKey(ContentType, related_name='+')
    assignee_id = models.PositiveIntegerField()
    assignee = GenericForeignKey('assignee_ct', 'assignee_id')

    role = models.ForeignKey(Role)

    class Meta:

        app_label = "djinn_auth"
        abstract = True

    def __unicode__(self):

        return "%s is %s" % (self.assignee, self.role)



from django.contrib import admin
from djinn_auth.models.role import Role
from djinn_auth.models.localrole import LocalRole
from djinn_auth.models.globalrole import GlobalRole
from djinn_auth.models.localpermission import LocalPermission


class RoleAdmin(admin.ModelAdmin):

    list_display = ['name']
    search_fields = ['name']
    ordering = ['-name']


class LocalRoleAdmin(admin.ModelAdmin):

    list_display = ['role', 'instance_ct', 'instance_id', 'assignee_id',
                    'assignee_ct']
    list_filter = ['role', 'instance_ct', 'assignee_ct']
    search_fields = ['role__name']


class GlobalRoleAdmin(admin.ModelAdmin):

    pass


class LocalPermissionAdmin(admin.ModelAdmin):

    pass


admin.site.register(Role, RoleAdmin)
admin.site.register(LocalRole, LocalRoleAdmin)
admin.site.register(GlobalRole, GlobalRoleAdmin)
admin.site.register(LocalPermission, LocalPermissionAdmin)



from django.core.exceptions import PermissionDenied


class PermissionProtectedMixin(object):

    """ Mixin to protect the view with a permission. If provided,
    the class may actually check against the type of call """

    permission = None

    def get_permission(self, request):

        if type(self.permission) == dict:
            return self.permission.get(request.method)
        else:
            return self.permission

    def check_permission(self, request):

        permission = self.get_permission(request)

        if not permission:
            return True

        try:
            obj = self.get_object()
        except:
            obj = None

        return request.user.has_perm(permission, obj=obj)

    def handle_unauthorized(self):

        raise PermissionDenied

    def dispatch(self, request, *args, **kwargs):

        if not self.check_permission(request):
            self.handle_unauthorized()

        return super(PermissionProtectedMixin, self).dispatch(
            request, *args, **kwargs)



from django.test.testcases import TestCase
from django.contrib.auth.models import User
from django.views.generic.base import View
from django.core.exceptions import PermissionDenied
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from djinn_auth.views.base import PermissionProtectedMixin
from djinn_auth.models.role import Role
from djinn_auth.utils import assign_local_role


class BaseView(PermissionProtectedMixin, View):

    def get(self, request, *args, **kwargs):

        return "GET OK"

    def post(self, request, *args, **kwargs):

        return "POST OK"


class OpenView(BaseView):

    pass


class ClosedView(BaseView):

    permission = "djinn_auth.view"


class SemiClosedView(BaseView):

    permission = {'GET': "djinn_auth.view"}


class ClosedDetailView(BaseView):

    permission = "djinn_auth.view"
    obj = None

    def get_object(self):

        return self.obj


class Request(object):

    def __init__(self, method, user):

        self.method = method
        self.user = user


class PermissionProtectedMixinTest(TestCase):

    def setUp(self):

        self.user = User.objects.create(username="bobdobalina")

        contenttype = ContentType.objects.get(app_label='contenttypes',
                                              model='contenttype')

        self.perm, created = Permission.objects.get_or_create(
            codename="view",
            content_type=contenttype, defaults={'name': 'View content'})

    def test_get(self):

        view = OpenView.as_view()

        self.assertEquals("GET OK", view(Request("GET", self.user)))

        view = ClosedView.as_view()

        try:
            view(Request("GET", self.user))
            self.fail("GET should be forbidden")
        except PermissionDenied:
            pass

        self.user.user_permissions.add(self.perm)

        self.assertEquals("GET OK", view(Request("GET", self.user)))

    def test_methods(self):

        view = SemiClosedView.as_view()

        self.assertEquals("POST OK", view(Request("POST", self.user)))

        try:
            view(Request("GET", self.user))
            self.fail("GET should be forbidden")
        except PermissionDenied:
            pass

    def test_object_view(self):

        thing = self.user

        view = ClosedDetailView.as_view(obj=thing)

        try:
            view(Request("GET", self.user))
            self.fail("GET should be forbidden")
        except PermissionDenied:
            pass

        # Let's have some fun and assign a local permission to self...
        #
        owner, created = Role.objects.get_or_create(name="owner")
        owner.add_permission(self.perm)

        assign_local_role(self.user, thing, owner)

        self.assertEquals("GET OK", view(Request("GET", self.user)))
